Attribute VB_Name = "PointGeometry"
Option Explicit

' Find the points nearest the upper left, upper right,
' lower left, and lower right corners.
Public Sub GetMinMaxCorners(ByRef iul As Integer, ByRef iur As Integer, ByRef ill As Integer, ByRef ilr As Integer)
Dim ul As Integer
Dim ur As Integer
Dim ll As Integer
Dim lr As Integer
Dim pt As Integer

    ' Start with the first point.
    iul = 1
    iur = 1
    ill = 1
    ilr = 1
    With m_Points(1)
        ul = -.X - .Y
        ur = .X - .Y
        ll = -.X + .Y
        lr = .X + .Y
    End With

    ' Search the other points.
    For pt = 2 To m_NumPoints
        With m_Points(pt)
            If ul < -.X - .Y Then
                iul = pt
                ul = -.X - .Y
            End If
            If ur < .X - .Y Then
                iur = pt
                ur = .X - .Y
            End If
            If ll < -.X + .Y Then
                ill = pt
                ll = -.X + .Y
            End If
            If lr < .X + .Y Then
                ilr = pt
                lr = .X + .Y
            End If
        End With
    Next pt
End Sub
' Find the box that fits in the corners.
Public Sub GetMinMaxBox(ByRef xmin As Single, ByRef xmax As Single, ByRef ymin As Single, ByRef ymax As Single)
Dim iul As Integer
Dim iur As Integer
Dim ill As Integer
Dim ilr As Integer

    ' Get the corners.
    GetMinMaxCorners iul, iur, ill, ilr

    ' Get the coordinates of a box inside this trapezoid.
    With m_Points(iul)
        xmin = .X
        ymin = .Y
    End With
    With m_Points(iur)
        xmax = .X
        If ymin < .Y Then ymin = .Y
    End With
    With m_Points(ilr)
        If xmax > .X Then xmax = .X
        ymax = .Y
    End With
    With m_Points(ill)
        If xmin < .X Then xmin = .X
        If ymax > .Y Then ymax = .Y
    End With
End Sub
' Cull points out of the convex hull that lie inside the
' trapezoid defined by the vertices with smallest and
' largest X and Y coordinates.
Public Sub HullCull()
Dim xmin As Single
Dim xmax As Single
Dim ymin As Single
Dim ymax As Single
Dim min_i As Integer
Dim max_i As Integer
Dim cullable As Boolean
Dim temp_pt As POINTAPI

    ' Find a culling box.
    GetMinMaxBox xmin, xmax, ymin, ymax

    ' Move points that lie inside the box to
    ' the end of the m_Points array.
    ' Points that have not been checked lie between
    ' min_i and max_i inclusive.
    min_i = 1
    max_i = m_NumPoints
    Do While min_i <= max_i
        ' See if point min_i is in the box.
        With m_Points(min_i)
            cullable = _
                (.X > xmin) And (.X < xmax) And _
                (.Y > ymin) And (.Y < ymax)
        End With

        If cullable Then
            ' Cull this point by swapping it to the end.
            temp_pt = m_Points(min_i)
            m_Points(min_i) = m_Points(max_i)
            m_Points(max_i) = temp_pt
            max_i = max_i - 1
        Else
            ' Don't cull this point.
            min_i = min_i + 1
        End If
    Loop
    m_NumEligiblePoints = max_i
End Sub
' Make a convex hull.
Public Function MakeConvexHull() As Long
Dim points_examined As Long
Dim pt As Integer
Dim best_pt As Single
Dim temp_pt As POINTAPI
Dim sweep_angle As Single
Dim best_angle As Single
Dim X As Single
Dim Y As Single
Dim test_angle As Single

#If RANDOMIZE_POINTS Then
    ' For debugging.
    ' Randomize the array of points.
    Dim i As Integer
    Dim j As Integer
    Randomize
    For i = 1 To m_NumEligiblePoints
        j = Int(Rnd * (m_NumEligiblePoints - i + 1)) + i
        temp_pt = m_Points(j)
        m_Points(j) = m_Points(i)
        m_Points(i) = temp_pt
    Next i
#End If

    ' Find the point with the smallest Y coordinate.
    best_pt = 1
    For pt = 2 To m_NumEligiblePoints
#Const USE_SMALLEST_X = True
#If USE_SMALLEST_X Then
        ' This version takes the one with the smallest X coordinate.
        If (m_Points(pt).Y < m_Points(best_pt).Y) Or _
           ((m_Points(pt).Y = m_Points(best_pt).Y) And _
            (m_Points(pt).X < m_Points(best_pt).X)) _
        Then
            best_pt = pt
        End If
#Else
        If (m_Points(pt).Y < m_Points(best_pt).Y) Then best_pt = pt
#End If
    Next pt

    ' Add this point to the convex hull.
    m_NumHullPoints = 1
    temp_pt = m_Points(m_NumHullPoints)
    m_Points(m_NumHullPoints) = m_Points(best_pt)
    m_Points(best_pt) = temp_pt

    ' Start wrapping up the other points.
    sweep_angle = 0
    Do
        ' Find the point with smallest AngleValue
        ' from point m_Points(m_NumHullPoints).
        With m_Points(m_NumHullPoints)
            X = .X
            Y = .Y
        End With
        best_pt = 0
        best_angle = 360

        ' Search the rest of the points.
        For pt = m_NumHullPoints + 1 To m_NumEligiblePoints
            With m_Points(pt)
                test_angle = AngleValue(X, Y, .X, .Y)
                points_examined = points_examined + 1
            End With
            If (test_angle >= sweep_angle) And _
               (best_angle > test_angle) _
            Then
                best_angle = test_angle
                best_pt = pt
            End If
        Next pt

        ' If we didn't find a point to use, we're done.
#If Not USE_SMALLEST_X Then
        If best_pt = 0 Then Exit Do
#End If

        ' See if the first point is better.
        ' If so, we are done.
        With m_Points(1)
            test_angle = AngleValue(X, Y, .X, .Y)
            points_examined = points_examined + 1
        End With
        If (test_angle >= sweep_angle) And _
           (best_angle >= test_angle) _
        Then
            ' The first point is better. We're done.
            Exit Do
        End If

        ' Uncomment these two lines for debugging.
        'EditorForm.DrawConvexHull
        'Stop

        ' Add the point to the convex hull.
        m_NumHullPoints = m_NumHullPoints + 1
        temp_pt = m_Points(m_NumHullPoints)
        m_Points(m_NumHullPoints) = m_Points(best_pt)
        m_Points(best_pt) = temp_pt

        sweep_angle = best_angle
    Loop

    MakeConvexHull = points_examined
End Function
' Return a number that gives the ordering of angles
' WRST horizontal from the point (x1, y1) to (x2, y2).
' In other words, AngleValue(x1, y1, x2, y2) is not
' the angle, but if:
'   Angle(x1, y1, x2, y2) > Angle(x1, y1, x2, y2)
' then
'   AngleValue(x1, y1, x2, y2) > AngleValue(x1, y1, x2, y2)
' this angle is greater than the angle for another set
' of points, then this number for
'
' This function is dy / (dy + dx).
Public Function AngleValue(ByVal x1 As Single, ByVal y1 As Single, ByVal x2 As Single, ByVal y2 As Single) As Single
Dim dx As Single
Dim dy As Single
Dim ax As Single
Dim ay As Single
Dim t As Single

    dx = x2 - x1
    ax = Abs(dx)
    dy = y2 - y1
    ay = Abs(dy)
    If ax + ay = 0 Then
        ' If the two points are the same, return 360.
        t = 360 / 9
    Else
        t = dy / (ax + ay)
    End If
    If dx < 0 Then
        t = 2 - t
    ElseIf dy < 0 Then
        t = 4 + t
    End If
    AngleValue = t * 90
End Function

